/*
* Title			:   CubeFSTrack.java
* Package		:   service.cubefs
* Project		:   Cube Platform
* version               :   Alfa 0
* Description           :   Almacena un fragmento de archivo apto para insertarse 
*                           en el anillo.
*          
*                           		
* Created on            :   02/07/2007 
* Author		:   Victor Puertas Garcia.
* Copyright		:   (c) 2007, Victor Puertas Garcia. vpgugr@gmail.com 
* 					   
* 
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later  version.
* See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html)for more details.
* 
* HISTORIAL DE CAMBIOS. Formato: <fecha> <autor> <descripcion>
*
*
*/

package service.cubefs;

import rice.p2p.commonapi.Id;
import rice.p2p.past.ContentHashPastContent;
import rice.p2p.past.PastContent;
import rice.p2p.past.PastException;


/**
 * Contiene un fragmento de archivo listo para almacenarse en el anillo.<b>PERSISTENCIA: TOTAL</b>. El fragmento
 * es identificado por el trackID resultado de aplicar un algoritmo resumen al contenido (data). <br>
 * 
 * Atributos de un fragmento:<br>
 * - fileID.<br>
 * - datos.<br>
 * - offset.<br>
 * - Atributos.<br><br>
 * <B>Limitaciones</B><br>
 * - No posee atributos para gestionar los fragmentos en anillos privados. Para ello
 * es necesario almacenar el nodeID del que publica y el nombre del grupo.<br><br>
 *
 * @author victor Puertas.
 */
public class CubeFSTrack extends ContentHashPastContent {
    
    /**
     * Almacena el identificador de archivo al que pertenece el fragmento.
     */
     private FileID fileID;
    /**
     * Version del modelo de fragmento.
     */
     static private long versionID = 10000;
    /**
     * Atributos y privilegios del fragmento.
     * Atributos: privado, cifrado.
     * privilegios: permisos.
     */
     private byte attributes = 0; 
    /**
     * Desplazamiento del fragmento dentro del archivo original.
     */
     private long offset = 0;
    /**
     * Almacena el fragmento del archivo.
     */
     private byte data [];
     /**
      * Tipo de persistencia bajo la que se ampara el objeto:
      */
     private byte persistent = CubeFS_impl.PERSISTENT_TOTAL;
     /**
      * Contadore de persistencia: permite determinar la expiracion de un objeto en 
      * caso de la persistencia TEMPORAL.
      */
     private long expirationTime = 0;
    
    /**
     * Permite construir un fragmento de archivo con persistencia TOTAL.
     * @param trackID Identificador de fragmento proporcionado por PAST.
     * @param fileID Identificador de archivo asignado al fragmento.
     * @param privado Indica si el contenido del fragmento es de tipo privado.
     * @param offset Desplazamiento del fragmento en el archivo original.
     */
    public CubeFSTrack(Id trackID, FileID fileID, byte attributes, long offset, byte data [] ) {
        super(trackID);
        this.fileID = fileID;
        this.offset = offset;
        this.data = data;
        this.attributes = attributes;
        persistent = CubeFS_impl.PERSISTENT_TOTAL;
    }
    
       /**
     * Permite construir un fragmento de archivo con persistencia TEMPORAL.
     * @param trackID Identificador de fragmento proporcionado por PAST.
     * @param fileID Identificador de archivo asignado al fragmento.
     * @param privado Indica si el contenido del fragmento es de tipo privado.
     * @param offset Desplazamiento del fragmento en el archivo original.
     * @param expiration Tiempo de expiracion del objeto.
     */
    public CubeFSTrack(Id trackID, FileID fileID, byte attributes, long offset, byte data [], long expiration ) {
        super(trackID);
        this.fileID = fileID;
        this.offset = offset;
        this.data = data;
        this.attributes = attributes;
        expirationTime = expiration;
        persistent = CubeFS_impl.PERSISTENT_TEMPORAL;
    }
    
    /**
     * Metodo invocado por Past que determina las acciones a realizar tras insertar, si es posible, el objeto.
     */
    public PastContent checkInsert(Id id, PastContent existingContent) throws PastException {
    // El objeto ya existe en el nodo:
    if (existingContent != null) {
      throw new PastException("CUBEFS:0000");
    }
    
    // Hash key erroneo:
    if (!id.equals(getId())) {
      throw new PastException("CUBEFS:0001");
    }
    return this;
  }
    /**
     * Obtiene la version del formato del fragmento.
     */
    public long getVersion() { return versionID; }
    
    /**
     * Obtiene el identificador de archivo al que pertenece el fragmento.
     * @return Identificador de archivo.
     */
    public FileID GetFileID() { return fileID; }
    
    /**
     * Devuelve los atributos del fragmento.
     * @return Byte con la mascara de atributos.
     */
    public byte GetAttributes() { return attributes; }
    
    /**
     * Obtiene el desplazamiento en bytes del bloque de datos dentro del archivo original.
     * Permite hacer la reconstrucción de los datos.
     * @return Desplazamiento del bloque en el archivo original.
     */
    public long GetOffset() { return offset; }
    
    /**
     * Obtiene los bytes del fragmento de archivo.
     * @return Conjunto de bytes que componen el fragmento de archivo.
     */
    public byte [] GetData() { return data; }
    
    /**
     * Obtiene el tipo de persistencia bajo la que se almacena el objeto.
     */
    public byte sortPersistent() {return persistent;}
    
    /**
     * Recarga el contador de tiempo, impudiendo que el objeto desaparezca del nodo.
     */
    //public void refresh(long time) {expirationTime = time;}
    

    
    /**
     * Devuelve una cadena de texto con informacion del fragmento.
     */
    public String toString() {return "Track: "+ this.getId().toStringFull()+" FileID: "+fileID.toString();}
}
