/*
* Title			:   FileManager.java
* Package		:   service.cubefs
* Project		:   Cube Platform
* version               :   Alfa 0
* Description           :   Permite manegar archivos FS.
*                           		
* Created on            :   11/08/2007 
* Author		:   Victor Puertas Garcia.
* Copyright		:   (c) 2007, Victor Puertas Garcia. vpgugr@gmail.com 
* 					   
* 
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later  version.
* See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html)for more details.
* 
* HISTORIAL DE CAMBIOS. Formato: <fecha> <autor> <descripcion>
*
*/

package service.cubefs;

import cube.common.logger.LogZero;
import cube.common.logger.NoActiveLoggerException;
import java.io.EOFException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;
import service.cubefs.exception.BadIdentifier;

/**
 * Se encarga de la logica de los archivos FS. Mediante la clase se puede extraer 
 * toda la informacion del fichero, almacenar/ obtener fragmentos.<br><br>
 * Caracteristicas:<br>
 * Version del formato de fichero: 1000
 * @author victor Puertas Garcia.
 */
public class FileFS {
    
    //Logger de la clase:
    static private Logger mylog;
   
    //Marca del formato de fichero (Magic number):
    static private final String MAGIC_NUMBER = "FS07";
    
   
    
    //Flujo de datos asociado al fs:
    private RandomAccessFile fsStream;
    //Contiene el nombre del fichero FS:
    private String fs_file;
    //Indice actual de escritura del trackID:
    private int trackIndex = 0;
    
    //Informacion extraida de la cabecera del fichero:
    //fileID;
    private FileID fileID;
    //Longitud del fragmento:
    private int track_long = 0;
    //Privilegios del archivo:
    private byte privileges = 0;
    //Version del formato de archivo implementado.
    public final int VERSION_FILE = 1005;
    //Version del archivo cargado:
    private int version = -1;
    //Metadatos asociados al archivo:
    private MetadataFile mymetadata = null;
    
    //Longitud de la cabecera:
    private int head_size = 0;
    
    //Longitud en bytes de los enteros en Java:
    static private final byte SIZE_INTEGER = Integer.SIZE/8;
    
    
    /**
     * Constructor de la clase
     */
    public FileFS(String fs_file)  { this.fs_file =fs_file;}
    
    /**
     * Abre un fichero FS existente y extrae la informacion de cabecera.
     */
    synchronized public void Open() throws FileNotFoundException, IOException
    {
        try 
        {
            
            fsStream = new RandomAccessFile(fs_file,"rw");
        } 
        catch (FileNotFoundException ex) 
        {
            //WriteLog(Level.WARNING, "El archivo: "+fs_file+" no se encuentra.");
            throw new FileNotFoundException("Fichero "+fs_file+" no disponible.");
        }
        
        try { ReadFSfileHead(); } catch (BadIdentifier ex) {} 
    }
    
    /**
     * Cierra el stream de datos. <B>Es importante cerrar el archivo antes para no perder datos.</B>
     */
    synchronized public void Close()
    {
        try {
            fsStream.close();
        } catch (IOException ex) {
            System.err.println("Error tipo I/O durante el cierre de el archivo: "+fs_file+"\n");
        } catch(java.lang.NullPointerException ex){}
    }
    
    /**
     * Crea un nuevo fichero FS, escribiendo la cabecera.
     */
    synchronized public void Create(FileID fileID, int track_long, byte privileges, MetadataFile metadata ) throws FileNotFoundException, IOException
    {
        this.fileID = fileID;
        this.track_long = track_long;
        this.privileges = privileges;
        this.mymetadata = metadata;
        
        File archivo = new File(fs_file);
        try {
            archivo.createNewFile();
        } catch (IOException ex) {
            System.err.println("Error al crear el archivo: "+fs_file);
            throw new IOException();
        }
        try {
            
            fsStream = new RandomAccessFile(archivo,"rw");
        } catch (FileNotFoundException ex) {
           System.err.println("El archivo: "+fs_file+" ha sido recientemente creado pero no se encuentra.");
            throw new FileNotFoundException("Fichero "+fs_file+" no disponible.");
        }
        try {
            CreateFSfileHead();
        } catch (IOException ex) {
            throw new IOException();
        }
    }
    
    /**
     * Extrae el trackID indicado en el indice.
     * @param index Indice, comenzando desde cero.
     */
    synchronized public byte [] GetTrackID(int index) throws EOFException, IOException 
    { 
        if(head_size < 29) throw new IOException("No se puede obtener el trackID indexado: Cabecera corrupta.");
        fsStream.seek((head_size)+(20*index)); 
        byte theID [] = new byte[20];
        fsStream.readFully(theID);
        return theID;
    }
    /**
     * Escribe el trackID en el archivo FS, en la posición del puntero actual de escritura 
     * de dicho archivo.
     */
    synchronized public void SetTrackID(byte [] trackID) throws IOException
    {
        SetTrackID(trackID,trackIndex);
        trackIndex++;
    }
    
    /**
     * Escribe el trackID en la posicion indicada por el indice.
     */
    synchronized public void SetTrackID(byte [] trackID, int index) throws IOException
    {
        if(head_size < 29) throw new IOException("No se puede obtener el trackID indexado: Cabecera corrupta.");
        fsStream.seek((head_size)+(20*index)); 
        fsStream.write(trackID);
    }
    
    /**
     * Modifica el indice de escritura.
     */
    synchronized public void SetIndex(int index) { trackIndex = index; }
    
    
    /**
     * Establece el puntero de lectura/escritura del fichero.
     * @param offset Desplazamiento en bytes. Comienza desde cero.
     */
    synchronized public void Seek(long offset) throws IOException { fsStream.seek(offset); }
    
    /**
     * Longitud actual del archivo FS.
     */
    synchronized public long GetFsSize() throws IOException { return fsStream.length(); }
    
    /**
     * Obtienele la longitud de cabecera. Se debe invocar tras llamar a Open().
     * @return Cero si no se ha leido la cabecera.
     */
    public int GetHeadSize() { return head_size; }
    
    /**
     * Permite consultar el numero de trackID albergados en el fichero.
     * @return Numero de trackID. -1 si ocurre algun error.
     */
    public int GetTotalTrack() {
        try {return (int)(fsStream.length()-head_size)/20;
        } catch (IOException ex) {
            System.err.println("Error tipo I/O al obtener la longitud del fichero: " + fs_file);
        } 
        return -1;
    }
    
    /**
     * Obtiene el identificador de archivo.
     */
    public FileID Get_info_FileID(){ return fileID; }
    
    /**
     * Obtiene la longitud del fragmento.
     */
    public int Get_info_TrackLong(){ return track_long; }
    /**
     * Obtiene los privilegios de archivos.
     */
    public byte Get_info_privileges(){ return privileges; }
    /**
     * Obtiene los metadatos.
     */
    public MetadataFile Get_info_metadata(){ return mymetadata; }
    
    /**
     * Devuelve un String con la informacion formateada del fichero en texto plano.
     */
    public String toString()
    {
        String FS_name = "FS file name: "+fs_file+"\n";
        String FS_ID = "File identifier: "+fileID.toString()+"\n";
        String date_file = "Created at "+  mymetadata.GetDate()+"\n";
        String cipher_data = "Cipher: ";
            if(Attribute.iscipher(privileges)) cipher_data+="yes\n"; else cipher_data+="no\n";
        String private_data = "Private: ";
            if(Attribute.isprivate(privileges)) private_data+="yes\n"; else private_data+="no\n";
        String privileges_file = "Privileges: ";
            if(Attribute.U_erase(privileges)) privileges_file+="E"; else privileges_file+="-";
            if(Attribute.U_download(privileges)) privileges_file+="D"; else privileges_file+="-";
            if(Attribute.G_erase(privileges)) privileges_file+="E"; else privileges_file+="-";
            if(Attribute.G_download(privileges)) privileges_file+="D"; else privileges_file+="-";
            if(Attribute.O_erase(privileges)) privileges_file+="E"; else privileges_file+="-";
            if(Attribute.O_download(privileges)) privileges_file+="D"; else privileges_file+="-";
            privileges_file+="\n";
            String number_track = "Total track identifier in this file: ";        
            number_track+=GetTotalTrack()+"\n";
       String version = "Version of file format: "+this.version+"\n";
       String author = "Author: "+mymetadata.GetAuthor()+"\n";
       String copyright = "Copyright: "+mymetadata.GetCopyright()+"\n";
       
       return FS_name+FS_ID+date_file+cipher_data+private_data+privileges_file+number_track+version+author+copyright;
               
    }
    
    /**
     * Devuelve la informacion del fichero FS en formato HTML.
     */
    public String toHtml()
    {
        String FS_name = "<b>FS file name:</b> "+fs_file+"<br>";
        String FS_ID = "<b>File identifier: </b>"+fileID.toString()+"<br>";
        String date_file = "<b>Created at</b> "+  mymetadata.GetDate()+"<br>";
        String cipher_data = "<b>Cipher:</b> ";
            if(Attribute.iscipher(privileges)) cipher_data+="yes<br>"; else cipher_data+="no<br>";
        String private_data = "<b>Private:</b> ";
            if(Attribute.isprivate(privileges)) private_data+="yes<br>"; else private_data+="no<br>";
        String privileges_file = "<b>Privileges:</b> ";
            if(Attribute.U_erase(privileges)) privileges_file+="E"; else privileges_file+="-";
            if(Attribute.U_download(privileges)) privileges_file+="D"; else privileges_file+="-";
            if(Attribute.G_erase(privileges)) privileges_file+="E"; else privileges_file+="-";
            if(Attribute.G_download(privileges)) privileges_file+="D"; else privileges_file+="-";
            if(Attribute.O_erase(privileges)) privileges_file+="E"; else privileges_file+="-";
            if(Attribute.O_download(privileges)) privileges_file+="D"; else privileges_file+="-";
            privileges_file+="<br>";
            String number_track = "<b>Total track identifier in this file:</b> ";        
            number_track+=GetTotalTrack()+"<br>";
       String version = "<b>Version of file format:</b> "+this.version+"<br>";
       String author = "<b>Author:</b> "+mymetadata.GetAuthor()+"<br>";
       String copyright = "<b>Copyright:</b> "+mymetadata.GetCopyright()+"<br>";
      
       return "<html><body>"+FS_name+FS_ID+date_file+cipher_data+private_data+privileges_file+number_track+version+author+copyright+"</body><html>";        
    }
    
    
     /**
     * Realiza una lectura de la cabecera del fichero .fs y extrae la informacion.
     * @throws service.cubefs.exception.BadIdentifier El fileID obtenido no es correcto. No cumple con el tamaño adecuado.
     * @throws java.io.IOException Error tipo I/O al leer la cabecera.
     */
    private void ReadFSfileHead() throws IOException, BadIdentifier
    {
        /*=================================================
         *             LEER LA CABECERA
         *=================================================
         */     
        try {
            fsStream.seek(0);
            head_size = 0;
            
             
            byte aux [];
            int longitud = 0;
            
            //Numero magico:
            aux = new byte [MAGIC_NUMBER.length()];
            fsStream.readFully(aux);
            String marca = new String(aux);
            head_size+=aux.length;
            if(!marca.equals(MAGIC_NUMBER))
                throw new IOException("Formato erroneo.");
            
            //FileID:
            aux = new byte [20];
            fsStream.readFully(aux);
            fileID = new FileID(aux);
            head_size+=20;
            
            //Longitud del fragmento:
            track_long = fsStream.readInt();
            head_size+=SIZE_INTEGER;
            
            //Mascara: caracteristicas + permisos del archivo:
            privileges = fsStream.readByte();
            head_size++;
            
            //Comprobamos la version del formato. Si es superior a la actual, salimos.
            version = fsStream.readInt();
            head_size+=SIZE_INTEGER;
            if(version!=1005) throw new IOException("Invalid file format");
            switch(version)
            {
                case 1005:  //Version de Dic-2007
                       
                        //Metadatos:
                            mymetadata = new MetadataFile();
                            Date date = null;

                            //FECHA:             
                            if((longitud = fsStream.readByte()) != 0)
                            {
                                aux = new byte [longitud];
                                fsStream.readFully(aux);
                                head_size += longitud+1;
                                try { date = new SimpleDateFormat("MM/dd/yy").parse(new String(aux)); }
                                catch (ParseException ex) {
                                   //System.err.println("Error al obtener la fecha del fichero FS.");
                                }
                                mymetadata.SetDate(new Date());
                            }

                            //AUTOR:
                            if((longitud = fsStream.readInt()) != 0)
                            {
                                aux = new byte [longitud];
                                fsStream.readFully(aux);
                                head_size += longitud+SIZE_INTEGER;
                                mymetadata.SetAuthor(new String(aux));
                            }

                            //COPYRIGHT:
                            if((longitud = fsStream.readInt()) != 0)
                            {
                                aux = new byte [longitud];
                                fsStream.readFully(aux);
                                head_size += longitud+SIZE_INTEGER;
                                mymetadata.SetCopyright(new String(aux));
                            }

                            //LICENCIA:
                            if((longitud = fsStream.readInt()) != 0)
                            {
                                aux = new byte [longitud];
                                fsStream.readFully(aux);
                                head_size += longitud+SIZE_INTEGER;
                                mymetadata.SetLicense(new String(aux));
                            }
                            
                            //DESCRIPCION:
                            if((longitud = fsStream.readInt()) != 0)
                            {
                                aux = new byte [longitud];
                                fsStream.readFully(aux);
                                head_size += longitud+SIZE_INTEGER;
                                mymetadata.SetDescription(new String(aux));
                            }
                            break;
            }    
        } catch (IOException ex) {
           System.err.println("Error tipo I/O al leer la cabecera en el fichero: " + fs_file);
           System.err.println(ex.toString());
           throw new IOException();
        }
        trackIndex = GetTotalTrack()+1;
    }
    
    /**
     * Crea la cabecera del fichero FS. y la deja lista para introducir
     * los TrackID.
     * @throws java.io.IOException Error tipo I/O al escribir la cabecera.
     */
    private void CreateFSfileHead() throws IOException
    {
        /*=================================================
         *             ESCRIBIR LA CABECERA
         *=================================================
         */
        try {
            ;
            fsStream.seek(0);
            head_size = 0;
            
            //Numero magico:
            fsStream.write(MAGIC_NUMBER.getBytes());
            head_size+=MAGIC_NUMBER.getBytes().length;
            
            //FileID:
            fsStream.write(fileID.GetBytes());
            head_size+=fileID.GetSize();
            //Longitud del fragmento:
            fsStream.writeInt(track_long);
            head_size+=SIZE_INTEGER;
            
            //Mascara: caracteristicas + permisos del archivo:
            fsStream.write(privileges);
            head_size++;
            //Versión del formato de archivo:
            fsStream.writeInt(VERSION_FILE);
            head_size+=SIZE_INTEGER;
            //Metadatos:
                //FECHA:
                if(mymetadata.GetDate() == null) {fsStream.write(0);}
                else{
                fsStream.write((byte) mymetadata.GetDate().length());
                fsStream.writeBytes(mymetadata.GetDate());}
                head_size+=mymetadata.GetDate().length()+1;
            
                //AUTOR:
                if(mymetadata.GetAuthor() == null) {fsStream.writeInt(0);}
                else{
                fsStream.writeInt(mymetadata.GetAuthor().length());
                fsStream.writeBytes(mymetadata.GetAuthor());}
                head_size+=mymetadata.GetAuthor().length()+SIZE_INTEGER;
            
                //COPYRIGHT:
                if(mymetadata.GetCopyright() == null) {fsStream.writeInt(0);}
                else{
                fsStream.writeInt(mymetadata.GetCopyright().length());
                fsStream.writeBytes(mymetadata.GetCopyright());}
                head_size+=mymetadata.GetCopyright().length()+SIZE_INTEGER;
            
                //LICENCIA:
                if(mymetadata.GetLicense() == null) {fsStream.writeInt(0);}
                else{
                fsStream.writeInt(mymetadata.GetLicense().length());
                fsStream.writeBytes(mymetadata.GetLicense());}
                head_size+=mymetadata.GetLicense().length()+SIZE_INTEGER;
                
                //DESCRIPCION:
                if(mymetadata.GetDescription() == null) {fsStream.writeInt(0);}
                else{
                fsStream.writeInt(mymetadata.GetDescription().length());
                fsStream.writeBytes(mymetadata.GetDescription());}
                head_size+=mymetadata.GetDescription().length()+SIZE_INTEGER;
                
                
        } catch (IOException ex) {
           System.err.println("Error tipo I/O al crear la cabecera del fichero FS: " + fs_file);
           throw new IOException();
        }        
    }
    
   
}
