/*
* Title			:   Task.java
* Package		:   service.cubefs
* Project		:   Cube Platform
* version               :   Alfa 0
* Description           :   Gestion y control de tareas en Cube File Service.
*                           		
* Created on            :   11/10/2007 
* Author		:   Victor Puertas Garcia.
* Copyright		:   (c) 2007, Victor Puertas Garcia. vpgugr@gmail.com 
* 					   
* 
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later  version.
* See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html)for more details.
* 
* HISTORIAL DE CAMBIOS. Formato: <fecha> <autor> <descripcion>
*
*
*/

package service.cubefs;

import java.util.HashMap;
import java.util.Iterator;

/**
 * Se encarga de almacenar los FileManager de los trabajos activos y determina el numero
 * maximo de operaciones que se pueden realizar en un momento dado, entendiendose por estas
 * subida o bajada de fragmentos. El concepto de trabajo hace referencia a un archivo que se
 * esta procesando, bien para subirlo o para extraerlo.
 * @author victor Puertas.
 */
public class Task {
   
    /**
     * Numero máximo de fragmentos que se pueden 
     * descargar de forma concurrente. Entre todos los
     * File Manager no pueden pedir mas de <maxConcurrentFragment> 
     * fragmentos. Esta variable es el valor limite de maxConcurrentTracksDown.
     */
    private int maxConcurrentTracksDown = 7;
   
    /**
     * Numero máximo de fragmentos que se pueden 
     * subir de forma concurrente. Entre todos los
     * File Manager no pueden subir mas de <maxConcurrentTracksUp> 
     * fragmentos. Esta variable es el valor limite de concurrentTrackUp.
     */
    private int maxConcurrentTracksUp = 5;
    /**
     * Sincronizador para las subidas:
     */
    private BoxSyncr syncrUp = new BoxSyncr();
    /**
     * Indica cuando se llego al maximo de la cola de subida.
     */
    private boolean trackUpQueueMax = false;
    /**
     * Numero maximo de FileManager activos en un instante temporal.
     */
    private int maxConcurrentFileManager = 16;
    /**
     * Numero de fragmentos que hay descargandose en un 
     * instante temporal.
     */
    private Integer concurrentTrackDown = 0;
    /**
     * Sincronizador para las bajadas:
     */
    private BoxSyncr syncrDown = new BoxSyncr();
     /**
     * Numero de fragmentos que hay descargandose en in 
     * instante temporal.
     */
    private Integer concurrentTrackUp = 0;
    
    
    /**
     * Permite ver cuando invocar al recolector de basura
     * para evitar los out of memory.
     */
    private int contador = 0;
    
    /**
     * Almacena los FileManager activos.
     */
    private HashMap<String,FileManager> almacen_file_manager;
    /**
     * Almacena el numero de fragmentos por FileManager.
     */
    private HashMap<String,Integer> trackPerFile = new HashMap(); 
    /**
     * Numero de FileManager activos en un instante de tiempo.
     */
    private int FileManagerActive = 1;
    
    /** Constructor */
    public Task() 
    {
        almacen_file_manager = new HashMap(20);
    }
    
    synchronized public void close()
    {
        trackPerFile.clear();
        almacen_file_manager.clear();
        concurrentTrackDown = 0;
        concurrentTrackUp = 0;
        contador = 0;
    }
    
    /**
     * Añade a la cola de bajada una nueva peticion de fragmento.
     */
    public void addTrackDown(String fileID)
    {
            synchronized(syncrDown)
            {
                if(concurrentTrackDown==maxConcurrentTracksDown)
                {      
                   syncrDown.Stop();
                }
                concurrentTrackDown++;

                try
                {
                    Integer value = trackPerFile.get(fileID);
                    value++;
                    trackPerFile.put(fileID,value);

                }
                catch (NullPointerException ex)
                {
                  System.err.println("Task addTrackDown: No existe una tarea asociada con el fileID: "+fileID);  
                }
            }
    }
    
    /**
     * Añade la cola de subida una nueva peticion de fragmento.
     */
    public void addTrackUp(String fileID)
    {   
        synchronized(syncrUp)
        {
            if(concurrentTrackUp==maxConcurrentTracksUp || trackUpQueueMax )
            {    
               trackUpQueueMax = true; 
               syncrUp.Stop();
            }
            concurrentTrackUp++;
            
            contador = ++contador%40;
            if(contador == 0) System.gc();
             
            try
            {
                Integer value = trackPerFile.get(fileID);
                value++;
                trackPerFile.put(fileID,value);
                
            }
            catch (NullPointerException ex)
            {
              System.err.println("Task addTrackUp: No existe una tarea asociada con el fileID: "+fileID);  
            }       
        }
    }
    
    /**
     * Elimina de la cola de bajada una peticion de fragmento .
     */
    public void subTrackDown(String fileID)
    { 
            synchronized(syncrDown)
            {
                concurrentTrackDown--;            
                syncrDown.Go();

                try
                {
                    Integer value = trackPerFile.get(fileID);
                    value--;
                    trackPerFile.put(fileID,value);
                }
                catch (NullPointerException ex)
                {
                  System.err.println("Task subTrackDown: No existe una tarea asociada con el fileID: "+fileID);  
                }
            }
    }
    
    /**
     * Elimina de la cola de subida una peticion de fragmento.
     */
    public void subTrackUp(String fileID)
    {
        synchronized(syncrUp)
        {
            concurrentTrackUp--;
            
            try
            {
                Integer value = trackPerFile.get(fileID);
                value--;
                trackPerFile.put(fileID,value);
            }
            catch (NullPointerException ex)
            {
              System.err.println("Task subTrackDown: No existe una tarea asociada con el fileID: "+fileID);  
            }    
            if(concurrentTrackUp==0) {trackUpQueueMax = false; syncrUp.Go();}
        }
    }
  
    /**
     * Obtiene el numero de fragmentos que hay en la cola correspondiente (subida o bajada)
     */
    synchronized public int getTrackInQueue(String fileID)
    {
        Integer value = trackPerFile.get(fileID);
        if(value != null)
            return value;
        return 0;
    }
    /**
     * Registra una nueva tarea.
     */
    public void addFileManager(FileManager fm)
    {
        synchronized(almacen_file_manager)
        {
            if(!almacen_file_manager.containsKey(fm.GetFileID().toString()))
                FileManagerActive++;   
           
            
            almacen_file_manager.put(fm.GetFileID().toString(),fm);
            trackPerFile.put(fm.GetFileID().toString(),0);
        }
    }
    
    /**
     * Elimina una tarea. Esta llamada eliminara por completo
     * la tarea. Se debe estar seguro cuando se invoca.
     * @param fileID asociado a la tarea.
     */
    public void subFileManager(String fileID)
    {
        synchronized(almacen_file_manager)
        {
            if(almacen_file_manager.remove(fileID)!=null)
            {
                FileManagerActive--;
                trackPerFile.remove(fileID);
            }
        }
    }
    
    /**
     * Devuelve una referencia del del FileManager asociado con el fileID
     * pasado como argumento.
     */
    public FileManager getFileManager(String fileID)
    {
        synchronized(almacen_file_manager)
        {
            return almacen_file_manager.get(fileID);
        }
    }
    
    /**
     * Devuelce un iterador con las referencias de los FileManager activos.
     */
    public Iterator<FileManager> getFileManagers()
    {
        synchronized(almacen_file_manager)
        {
             return almacen_file_manager.values().iterator();
        }       
    }
    
}
