/*
* Title			:   TrackManager.java
* Package		:   service.cubefs
* Project		:   Cube Platform
* version               :   Alfa 0
* Description           :   Se engarga de obtener y procesar los fragmentos de un archivo. Se crea 
*                           por cada archivo.
*          
*                           		
* Created on            :   28/05/2007 
* Author		:   Victor Puertas Garcia.
* Copyright		:   (c) 2007, Victor Puertas Garcia. vpgugr@gmail.com 
* 					   
* 
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later  version.
* See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html)for more details.
* 
* HISTORIAL DE CAMBIOS. Formato: <fecha> <autor> <descripcion>
*
*
*/


package service.cubefs;

import cube.common.logger.LogZero;
import cube.common.logger.NoActiveLoggerException;
import cube.common.postman.Postman;
import cube.common.security.Cryptography;
import java.io.EOFException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.crypto.SecretKey;
import rice.pastry.commonapi.PastryIdFactory;
import service.cubefs.exception.KeyCipherNotFound;

/**
 * Permite fragmentar y recomponer archivos. En el proceso de fragmentacion, procesa los
 * datos y los encapsula en un objeto tipo <B>ContentHashPastContent</B> para ser insertados
 * en el anillo. En el proceso de reconstruccion permite recuperar el archivo original.
 * El TrackManager no lleva un cotrol sobre los fragmentos creados o escritos. Dicha tarea
 * es encomendada al <B>FileManager</B> El uso de la clase es sencillo; Empleamos el
 * constructor para indicarle las opciones de fragmentación:
 * <B>Nombre del archivo</B>,<B> fileID</B>,<B> Clave de cifrado</B>,<B> si el fragmento es o no privado</B>
 * No es necesario cifrar el contenido, si bien por segurida, se emplea este metodo. 
 * El algoritmo de cifrado se puede obtener consultando la documentacion del Cube Cryptography Engine en el 
 * apartado de cifrado en bloques.
 * @author victor Puertas.
 */
public class TrackManager {
    /**
     * Acceso en modo lectura.
     */
    static public final String MODE_READ = "r";
    /**
     * Acceso en modo lectura y escritura. Si el archivo no existe <B>NO LO CREA.</B>
     */
    static public final String MODE_READ_WRITE = "rw"; 
    //Almacena el archivo sobre el que se extraeran/escriben los fragmentos.
    private File myfile;
    //Identificador asignado al archivo.
    private FileID fileID;
   
    //Privilegios y atributos del archivo:
    private byte attributes = 0; 
    //Longitud del fragmento en bytes:
    private int track_long;
    //Modo de trabajo seleccionado:
    private String mode;
    
    //Stream asociado al archivo sobre el que se extraeran/escriben los fragmentos.
    private RandomAccessFile theFile;
    //Buffer de datos para R/W. Tamaño longitud del fragmento: 
    //private byte buffer [];
    //Tamaño del ultimo bloque:
    private int size_last_Block = 0;
    
    //Indica si el TrackManager se ha creado con exito:
    private boolean activo = false;
    
    //Indica el numero de Fragmentos del archivo:
    private long number_track = 0;
    
    //Clave de cifrado empleada en los bloques:
    private SecretKey clave = null;
    
    //Generador de Identificadores de Pastry:
    private PastryIdFactory IDFactory;
    
    //Tipo de persistencia:
    private byte persistent = service.cubefs.CubeFS_impl.PERSISTENT_TOTAL;
   
    //CONTADORES:
    //Numero de fragmentos solicitados:
    private int counter_track = 0;
   
    
    /**
     * Constructor que permite crear una instancia del Track Manager.
     * @param IDFactory Generador de identificadores del anillo.
     * @param file Ruta completa o relativa del archivo sobre el que se montara el Track Manager. En caso de la descarga de datos,
     * el no se crea el archivo destino, debe existir previamente.
     * De este se obtendran los fragmentos necesarios.
     * @param fileID Identificador de archivo con el que se asociaran los fragmentos.
     * @param track_long Tamaño del fragmento de archivo.
     * @param mode Indica el modo de trabajo del TrackManager: lectura, escritura.
     * @param persistent Tipo de persistencia.
     * @throws java.io.FileNotFoundException El archivo pasado como argumento no se encuentra.
     * @throws java.lang.IllegalArgumentException El mode de trabajo pasado como argumento no es valido. Utiliza las constantes que
     * indican el modo de acceso.
     * @throws java.lang.SecurityException El gestor de seguridad de Java deniega el acceso al fichero en el modo pasado como
     * argumento.

     */
    public TrackManager(PastryIdFactory IDFactory, byte attributes, String file, String mode, FileID fileID, int track_long, byte persistent) throws FileNotFoundException, SecurityException, IllegalArgumentException {
        myfile = new File(file);
        this.fileID = fileID;
        this.track_long = track_long;
        this.IDFactory = IDFactory;
        this.mode = mode;
        this.attributes = attributes;
        this.persistent = persistent;
       
        
        //Si el archivo no existe, lo indicamos pero no creamos:
        if(!myfile.exists()) { System.err.println("El TrackManager no encuentra el archivo sobre el que trabajar."); throw new FileNotFoundException("TrackManager.");}
        
        try {
            theFile = new RandomAccessFile(myfile, mode);
            
                number_track = myfile.length() / track_long;
                size_last_Block = (int) myfile.length() % track_long;
                if (size_last_Block > 0) number_track ++;
                
        
            activo=true;
        } 
        catch (SecurityException ex) {
            System.err.println("El gestor de seguridad de Java no permite el acceso al fichero " + myfile.getName() + " en modo: " + mode + ".");
            throw new SecurityException("TrackManager.");
        }
        catch (IllegalArgumentException ex) {
            System.err.println("El modo de acceso: " + mode + " no es valido. Revise la documentación del servicio.");
            throw new IllegalArgumentException("TrackManager.");
        }

    } 
    /**
     * Liberar los recursos empleados.
     */
    public void close()
    {
        try {
            theFile.close();
        } catch (IOException ex) {
            System.err.println("Track Manager - fileID: "+fileID.toString()+" Error al liberar recursos.");
        }
    }
    /**
     * Estable la clave de cifrado/descrifrado empleada en archivos con tal atributo
     * activado. <B>Este metodo se debe llamar antes de realizar cualquier lectura/escritura
     * de bloque. De lo contrario salta una excepcion. Esto es solo si en los atributos
     * se ha establecido que el archivo posee contenido cifrado.</B>
     * @param key Clave empleada.
     */
    public void SetKey(SecretKey key) {clave = key;}
    
    /**
     * Obtiene un fragmento de archivo preparado para ser insertado en el anillo. <b>Persistencia TOTAL</b>.
     * @return Null si ha ocurrido algun problema.
     * @throws service.cubefs.exception.KeyCipherNotFound En los atributos del archivo esta activo el flag de cifrado y no se ha proporcionado
     * la clave con la que trabajar.
     * @param index Numero de fragmento. Comienza desde 0.
     */
    public CubeFSTrack ReadTrack(int index) throws KeyCipherNotFound, InvalidKeyException
    {
        if(!activo) {  return null; }
        
        if (Attribute.iscipher(attributes) && clave==null){ throw new KeyCipherNotFound(); };
        
        if(index > number_track) { System.err.println("Error al obtener el fragmento. Indice fuera de rango."); return null; }
        
        try
        {
            //Desplazamiento del bloque dentro del archivo:
            long offset = index * track_long;

            counter_track++;

            //Caso: primero fragmento.
            if (index == 0 ) 
            { 
                //Ultimo fragmento:
                if (index == number_track-1) { return CreateBlock(offset, true); }
                return CreateBlock(0, false); 
            }

            //Caso: resto de los fragmentos.
            else
            { 
                //Ultimo fragmento:
                if (index == number_track-1) { return CreateBlock(offset, true); }
                return CreateBlock(offset, false); 
            }
        }
        catch(InvalidKeyException ex){System.err.println("La clave de cifrado no es valida."); throw new InvalidKeyException();}
        catch(NoSuchAlgorithmException ex) { System.err.println("No se encuentra el algoritmo de cifrado."); }
        catch (EOFException ex){/*WriteLog(Level.INFO, "Fin de fichero inesperado.");*/}
        catch (IOException ex){ System.err.println("Error tipo I/O en el TrackManager, fragmento: " + index + " archivo: " + myfile.getName()); }
        return null;   
    }
    
    /**
     * Obtiene los datos del fragmento y los situa en la posicion adecuada del archivo
     * de reconstruccion.<b>Persistencia TOTAL</b>.<br> Gracias a este metodo, y pasandole cada fragmento recuperado
     * se puede obtener el archivo original. <B>Es recomendable pasar los fragmentos en orden 
     * ascendente para ocupar el minimo tamaño durante la reconstruccion. ESTE METODO SOLO SE 
     * EMPLEA CON EL MODO <I>mode_read_write</I> DE TRABAJO</B>.
     * @param track Fragmento de archivo obtenido del anillo.
     * @return <B>true</B> si el fragmento se ha escrito correctamente. <B>false</B> de lo contrario.
     * La causa de que no se escriba el fragmento suelen ser errores de tipo I/O al trabajar
     * con el archivo de reconstrucción tales como: al obtener el tamaño del mismo, posicionarse...
     * Otra causa es no elegir bien el modo de trabajo: <B>mode_read_write</B> o <B>mode_read</B>.
     * @throws service.cubefs.exception.KeyCipherNotFound En los atributos del archivo esta activo el flag de cifrado y no se ha proporcionado
     * la clave con la que trabajar.
     */
    synchronized public boolean WriteTrack(CubeFSTrack track) throws KeyCipherNotFound, InvalidKeyException
    {
        if(mode != MODE_READ_WRITE) { System.err.println("Para reconstruir archivos debe elegir el modo: MODE_READ_WRITE de trabajo"); return false; }
        long size_of_file = 0;
        
        //Solo para cuando el contenido se debe cifrar:
        byte resultado_cipher [] = null;
            
        if (Attribute.iscipher(attributes) && clave==null) throw new KeyCipherNotFound();
        
        if (Attribute.iscipher(attributes))
        {
            try { Cryptography.UnCipherBytes(clave,track.GetData(),resultado_cipher); }
             catch (InvalidKeyException ex) { System.err.println("La clave de cifrado no es valida."); throw new InvalidKeyException(); }
             catch (NoSuchAlgorithmException ex) { System.err.println("No se encuentra el algoritmo de cifrado."); } 
        }
        
        try { 
                size_of_file = theFile.length(); 
                
                /*En este caso no es necesario controlar si es el ultimo fragmento debido a que se aumenta el archivo
                teniendo en cuenta el tamaño bruto de los datos almacenados en el track.*/
                if(size_of_file < track.GetOffset()+track.GetData().length)
                {
                    //Incrementamos la longitud del archivo:
                    long incremento = (track.GetOffset()-theFile.length())+track.GetData().length;
                    theFile.setLength(theFile.length()+incremento);                    
                }
                
                theFile.seek(track.GetOffset());
                
                if (Attribute.iscipher(attributes)) theFile.write(resultado_cipher);
                else theFile.write(track.GetData());
                    
                
        
        } catch (IOException ex) {
            System.err.println("Error tipo I/O al trabajar con el archivo de reconstruccion: " + myfile.getName());
            return false;
        }
        
       return true;
    }
        
    /**
     * Obtiene el numero total de fragmentos del archivo. Para ello realiza una simple
     * division y en caso de no ser exacta incrementa el resultado en uno para cubrir
     * los datos restantes.
     * @return 0 si ha ocurrido algun problema al crear el TrackManager.
     */
    public long TotalTracks() { return number_track; }
    
    /**
     * Devuelve el numero de fragmentos solicitados.
     * @return Numero TOTAL de fragmentos solicitados hasta la fecha.
     */
    public int CounterTrack() {return counter_track; }
    
    /**
     * Se encarga de crear un fragmento a partir de un desplazamiento dado.
     * @param offset Desplazamiento desde donde se comenzara a extraer los datos.
     * @param last_case Indica si se trabaja con el ultimo fragmento del archivo.
     * @return El fragmento listo para insertarse en el anillo.
     * @throws java.io.IOException Error I/O durante el procesado del fragmento.
     * @throws java.security.InvalidKeyException Se ha establecido el contenido como cifrado y la clave para cifrar no es validad.
     * Puede que no se haya establecido.
     */
    private CubeFSTrack CreateBlock(long offset, boolean last_case) throws EOFException, IOException, InvalidKeyException, NoSuchAlgorithmException
    { 
            //Solo para cuando el contenido se debe cifrar:
            byte resultado_cipher [] = null;

            byte bufferWork []= new byte[track_long];
         
            theFile.seek(offset);
             
            //Solo para el ultimo caso: Creamos un buffer a medida:
            if(last_case)
            {
                CubeFSTrack fragment = null;
                
                byte block_buffer [] = new byte [size_last_Block];

                theFile.readFully(block_buffer);
                if(Attribute.iscipher(attributes)) 
                { 
                    Cryptography.CipherBytes(clave,block_buffer,resultado_cipher);
                    fragment = new CubeFSTrack(IDFactory.buildId(Cryptography.digestSHA1(block_buffer)), fileID, attributes, offset, resultado_cipher);
                }
                else fragment = new CubeFSTrack(IDFactory.buildId(Cryptography.digestSHA1(block_buffer)), fileID, attributes, offset, block_buffer);
                 
                return fragment;
            }
            //------------------------

            theFile.readFully(bufferWork);
            if(Attribute.iscipher(attributes)) 
            {
                Cryptography.CipherBytes(clave,bufferWork,resultado_cipher);
                //return new CubeFSTrack(IDFactory.buildId(bufferWork), fileID, attributes, offset, resultado_cipher);
                
                return new CubeFSTrack(IDFactory.buildId(Cryptography.digestSHA1(bufferWork)), fileID, attributes, offset, resultado_cipher);
            }
            else
            {
                
                CubeFSTrack fragment = null;
                
                if(persistent==service.cubefs.CubeFS_impl.PERSISTENT_TOTAL)
                    //Al buildID se le pasa un resumen para que Pastry no genere ID repetidos. Esto es debido a la forma en que lo obtiene.
                    fragment =  new CubeFSTrack(IDFactory.buildId(Cryptography.digestSHA1(bufferWork)), fileID, attributes, offset, bufferWork);
                else
                    fragment =  new CubeFSTrack(IDFactory.buildId(Cryptography.digestSHA1(bufferWork)), fileID, attributes, offset, bufferWork,CubeFS_impl.TIME_OUT_DATA);
                           
                 return fragment;
            }
    } 
}
