/*
* Title			:   LogZero.java
* Package		:   cube.common.logger
* Project		:   Cube Platform
* version               :   Alfa 0
* Description           :   Clase raiz del espacio de nombres de los log de la plataforma Cube. Toda clase
*                           que necesite un log debe obtenerlo de LogZero, que será el que proporcione al 
*                           nuevo log el prefijo del espacio de nombres correspondiente. Todos los log de un
*                           servicio se engloban bajo el nombre de Logger System. 
*                           		
* Created on            :   3/05/2007 
* Author		:   Victor Puertas Garcia.
* Copyright		:   (c) 2007, Victor Puertas Garcia. vpgugr@gmail.com 
* 					   
* 
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later  version.
* See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html)for more details.
* 
* HISTORIAL DE CAMBIOS. Formato: <fecha> <autor> <descripcion>
*
*
*/

package cube.common.logger;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.logging.*;

/**
 * Es la raiz del namespace de las clases log usadas en un servicio. Para mas 
 * informacion acerca de sistema jerarquico logger de Java, ver la API 
 * java.util.logging.
 * 
 * FORMATO ESPACIO DE NOMBRES: cube_platform.<nombre_servicio>.<subnombre>
 * ZONA DE ALMACENAMIENTO DEL LOG: <cube_path>/log
 * NOMBRE DEL ARCHIVO LOG: messages.
 * ARHIVO DE CONFIGURACION DEL LOGGER: <cube_path>/conf/logger.conf
 * @author victor Puertas.
 *@todo Modificar el sistema logger para que cada servicio tenga un archivo log pripio.
 *@todo Revisar el metodo writelog para que sea soncronizado con el objeto log y no con la clase en la que se alberga.
 */
public class LogZero {
   
    //Raiz del espacio de nombres del logger:
    static private String namespace = "cube_platform";
    //LogZero: Root de los log de servicio:
    static private Logger logzero;
    //Indica si el sistema logger esta activado:
    static private boolean active=false;
    
    /**
     * Constructor. No es necesario. Todos los metodos son estaticos.
     */
    public LogZero() {}
  
    /**
     * Permite crear un logger en el espacio de nombres de la plataforma Cube.
     * @param nombre Nombre del log. Normalmente es el nombre de la clase en la que se usara
     * el log.
     * @return Logger perteneciente al namespace del servicio.
     * @throws cube.common.logger.NoActiveLoggerException Producida en los casos en que el Logger System no esta activo.
     * Esto se debe a que en el arhivo de configuracion se ha desactivado.
     */
    static public Logger CreateSubLogger(String nombre) throws NoActiveLoggerException
    {
        if(!Active()){throw new NoActiveLoggerException("Logger System desactivado.");}
        return logzero.getLogger("."+nombre);
    }
    
    /**
     * Inicializa el sistema Logger:
     * - Carga el archivo de configuracion.
     * - Carga los manejadores oportunos.
     * - Crea a LogZero.
     * @param path Lugar en el que se encuentra el archivo de configuracion del logger del
     * servicio. Normalmente sera: <cube_path>/conf/<nombre_Servicio>/conf_logger
     * @throws java.io.IOException El motivo de la excepción se encuentra en el mensaje que adjunta la misma.
     * Fuentes de la excepcion:
     * - Error de IO.
     * - Error de seguridad: no hay permisos suficientes para acceder al archivo.
     * - Archivo no encontrado.
     * @throws cube.common.logger.NoActiveLoggerException Se produce cuando en el archivo de configuracion el Logger system se desactiva.
     */
    static public void StartLogger(String path) throws IOException, NoActiveLoggerException
    {
        //cargar la configuracion del LogManager:
            try {  
                LogManager.getLogManager().readConfiguration(new FileInputStream(path));
            } catch (SecurityException ex) {
                throw new IOException("LogZero: El nivel de seguridad no permite la apertura del archivo de configuracion del logger.");
            } catch (FileNotFoundException ex) {
               throw new IOException("LogZero: Archivo de configuracion del logger no encontrado.");
            } catch (IOException ex) {
                throw new IOException("LogZero: Error de IO en la apertura del fichero de configuracion del logger.");
            }
            
        // Ver si el logger esta activo:
        String cadenaConf=LogManager.getLogManager().getProperty("java.util.logging.ActivateLogger");
        if(cadenaConf==null)
            throw new IOException("Archivo de configuracion del Logger System dañado. No se encuentra la cadena: java.util.logging.ActivateLogger");
       
        
        if(cadenaConf.equals("no"))
        { 
            throw new NoActiveLoggerException("Logger System desactivado desde el archivo de configuracion.");
        }
        else { SetActive(true); }    
            
        //Crea la raiz de los Log:
        logzero=Logger.getLogger(namespace);
        //Añadir los Manejadores, encargados de volcar la informacion del log:
        logzero.addHandler(new FileHandler());
        //Manejador para log remoto:
        cadenaConf=LogManager.getLogManager().getProperty("java.util.logging.ActivateRemoteLog");
        if(cadenaConf==null)
            throw new IOException("Archivo de configuracion del Logger System dañado. No se encuentra la cadena: java.util.logging.ActivateRemoteLog");

        if(cadenaConf.equals("yes"))
            logzero.addHandler(new FileHandler());       
    }
    
    /**
     * Recarga la configuracion del Logger System desde el archivo de configuracion
     * especificado en StartLogger(). Si se lanza este metodo antes que StartLogger()
     * se recarga la configuracion a partir del archivo por defecto incorpardo en Java.
     * @throws java.io.IOException Error IO al leer el archivo de configuracion.
     */
    static public void ReloadConfigure() throws IOException{LogManager.getLogManager().readConfiguration();}
    
    /**
     * Permite comprobar si el sistema logger esta activado.
     * @return Indica si el sistema logger esta ativo.
     */
    synchronized static public boolean Active(){return active;}    
    /**
     * Establece el estado del sistema logger de la plataforma.
     * @param valor Estado nuevo estado del sistema logger.
     */
    synchronized static private void SetActive(boolean valor){active=valor;}   
    
}
