/*
* Title			:   Ring.java
* Package		:   cube.common.pastry
* Project		:   Cube Platform
* version               :   Alfa 0
* Description           :   Funciones necesarias para conectar a anillo publicos/privados.
*                           		
* Created on            :   25/07/2007 
* Author		:   Victor Puertas Garcia.
* Copyright		:   (c) 2007, Victor Puertas Garcia. vpgugr@gmail.com 
* 					   
* 
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later  version.
* See the GNU General Public License (http://www.gnu.org/copyleft/gpl.html)for more details.
* 
* HISTORIAL DE CAMBIOS. Formato: <fecha> <autor> <descripcion>
*
*
*/

package cube.common.pastry;

import cube.common.logger.LogZero;
import cube.common.logger.NoActiveLoggerException;
import cube.common.postman.Postman;
import cube.common.util.Pair;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.*;
import java.util.logging.*;
import rice.environment.Environment;
import rice.pastry.*;
import rice.pastry.socket.SocketPastryNodeFactory;

/**
 * Permite la conexion con anillos. Esta clase solo se debe instanciar en el Postman.
 * @author victor Puertas.
 *@todo Acceso a anillos privados.
 */
public class Ring {
    
    /**
     * Acceso a un anillo de tipo publico.
     */
    static public final byte PUBLIC_RING = 0;
    /**
     * Acceso a un anillo de tipo privado.
     */
    static public final byte PRIVATE_RING = 1;
    
    //Logger de la clase:
    static private Logger mylog;
    //Nombre del logger:
    static private final String logger_name = "service.CubePlatform.Ring";
    
      static
    {
        //Iniciamos el log para la clase:
        try { mylog = LogZero.CreateSubLogger(logger_name); } catch (NoActiveLoggerException ex) {}
    }
      
    /**
     * Mapa tipo hash que almacena los pares <alias> <PastryNode>.<br>
     * Como clave se emplea el alias.
     */
    private HashMap <String, Pair> storeConnection;
    //Generador del identificador del nodo: nodeID
    private NodeIdFactory nidFactory;
   
    
    public Ring() 
    {
        
        //Mapa con las referencias a los objetos de conexion con el anillo:
        storeConnection = new HashMap( new Integer(Postman.getConfigParamater("cube.numHashRing")));  
    }
    
    
    /**
     * Conecta con un anillo. Para ello es necesario conocer la IP y el puerto de un nodo
     * participante del anillo. En caso de ser privado, necesitamos el certificado de nodo
     * firmado por el CA del anillo. Esto implica que nuestro nodo debe estar tambien en posesión
     * del certificado del CA.<br>
     * cada conexion posee un alias que la identifica, lo que permite obtener la referencia del nodo
     * de para poder conectar varios servicios.<br><br><br>
     * <b><strong>Caracteristicas actuales de la conexion:<strong></b><br>
     * <ul>
     *      <li>La identificacion es unica para todos los anillos en los que conecte.</li>
     *      <li>Posibilidad de conectar con varios anillos.</li>
     * </ul>
     * @param alias Alias asignado a la conexion.
     * @param certNode Certificado del nodo empleado en la conexion.
     * @param tipo Tipo de conexión: anillo publico o privado. Para ello emplear las constantes PUBLIC_RING y PRIVATE_RING.
     * @param bindport Puerto de la maquina local en el que se recibiran peticiones.
     * @param createRing Indica si se crea un anillo o se conecta a uno existente.
     * @param bootaddress Direccion IP:puerto del nodo que servira de enlace.
     * @param env Opciones del entorno.
     * @return Objeto tipo PastryNode, que representa al nodo. <B>NULL</B> Si el alias ya estaba introducido. Consultar el log.
     * @throws java.io.IOException Se ha producido un error durante la conexion con el nodo de enlace.
     */
    //public PastryNode JoinToRing(String alias, java.security.cert.Certificate certNode, byte tipo,  boolean createRing, int bindport, InetSocketAddress bootaddress, Environment env) throws IOException
    public PastryNode JoinToRing(Properties conf, java.security.cert.Certificate certNode, int bindport, InetSocketAddress bootaddress, Environment env) throws IOException
    {
        String alias, nameRing;
        boolean createRing = false;
        
        //Alias de la conexion:
        if((alias = conf.getProperty("ring.alias")) == null) throw new IOException("Bad-properties");
        if(storeConnection.containsKey(alias) ) return null;
        
        //¿Crear anillo?
        if(conf.getProperty("ring.create")!=null) createRing = true;
        
        //Nombre del anillo
        nameRing = conf.getProperty("ring.name");
        if(nameRing == null && createRing) throw new IOException("Bad-properties");
        
        //Generador de identificadores de nodo: Si no hay certificado lo generamos en base a cube.chain.
        if(certNode != null)
            nidFactory = new CubeNodeCertIdFactory(certNode);
        else
        {
            //Concatenamos la cadena aleatoria con el nombre del nodo:
            String cadena = Postman.getConfigParamater("cube.chain")+Postman.getNodeName();
            nidFactory = new CubeNodeChainIdFactory(cadena);
        }
        
        PastryNodeFactory factory = null;
        PastryNode node;
        
        try 
        {   
            factory = new SocketPastryNodeFactory(nidFactory,bindport, env);            
        } 
        catch (IOException ex) {
            System.err.println("Ring>> No se puede conectar con el nodo de enlace: "+bootaddress.toString()+" para el alias: "+alias);
            throw  new IOException();
        }
        
         NodeHandle bootHandle = ((SocketPastryNodeFactory)factory).getNodeHandle(bootaddress);
         
         if(bootHandle==null && !createRing) throw new IOException("Ring>>No se encuentra nodo destino.");
         
         //Construccion del nodo:
         node = factory.newNode(bootHandle);
         
          synchronized(node) 
         {
           while(!node.isReady() && !node.joinFailed()) 
           {
                // Esperamos por si esta ocuapado...
                try{node.wait(500);} catch(InterruptedException ex){}

               //Aborta si no puede conectar:
                if (node.joinFailed()) {
                  System.err.println("Error al enlazar con el anillo. Alias:"+alias+" Razones: "+node.joinFailedReason());  
                  throw new IOException(); 
                }
           }       
         }
         if(createRing) node.pushRingName(nameRing);
         
         System.out.println("Nombre del anillo: "+node.readRingName());
         System.out.println("Ring>> Politica NAT: "+node.getEnvironment().getParameters().getString("nat_search_policy"));
         
         //Almacenamos los objetos asociados al alias:
         storeConnection.put(alias, new Pair(certNode, node));
         return node;
    }
    /**
     * Permite contectar con un anillo empleado un generador de identificacion de nodo basado
     * en la cadena de bytes almacenada en cube.chain del fichero conf_cube.
     */
    //public PastryNode JoinToRing(String alias, byte tipo, boolean createRing, int bindport, InetSocketAddress bootaddress, Environment env) throws IOException
    public PastryNode JoinToRing(Properties conf, int bindport, InetSocketAddress bootaddress, Environment env) throws IOException
    {
        return JoinToRing(conf, null, bindport, bootaddress, env);
    }
    
    /**
     * Obtiene una referencia del objeto tipo PastryNode que representa al nodo en el anillo.
     * Para ello empleamos el alias asignado a la conexion.
     * @param alias Alias asignado a la conexion.
     * @return Objeto que representa al nodo o <B>NULL</B> si dicho alias no existe.
     */
    public PastryNode GetNodeRef(String alias) {return (PastryNode) storeConnection.get(alias).element2();}
    
    /**
     * Devuelve el identificador de nodo empleado en la conexion identificada por el alias.
     * @return null si no hay conexion con ese alias.
     */
    public Id NodeID(String alias) 
    { 
        PastryNode nodo = GetNodeRef(alias);
        if(nodo !=null) return (Id) nodo.getId();
        return null;
    }
    
    /**
     * Devuelve el certificado usado por el nodo.
     */
    public java.security.cert.Certificate getNodeCert(String alias) { return (java.security.cert.Certificate) storeConnection.get(alias).element1(); }
    
    /**
     * Devuelve el nombre del anillo al que esta asociado un nodo.
     * @return null si no hay conexion con ese alias.
     */
    public String getRingName(String alias) 
    {
        PastryNode nodo = GetNodeRef(alias);
        if(nodo !=null) return nodo.readRingName();
        return null; 
    }
    
    /**
     * Cierra la conexion con el anillo.
     * @param alias Alias de la conexion empleada.
     * @return TRUE La conexion se cerro con exito.<br>
     * @return FALSE El alias pasado no corresponde con ninguna conexion.
     */
    public boolean closeNode(String alias)
    { 
        try
        {
            PastryNode theNode = (PastryNode) storeConnection.get(alias).element2();
       
            theNode.destroy();
            storeConnection.remove(alias);
            System.out.println("Ring>> Cerrada la conexión con el anillo.");
            return true;
        }
        catch(NullPointerException ex){}
        return false;
    }
    
    /**
     * Facilita la escritura de mensajes en el Logger solo si este está activo.
     * 
     */
    private static void WriteLog(Level nivel,String mensaje) {if (LogZero.Active()){ synchronized(mylog){mylog.log(nivel,mensaje);} }}
}