/*
 * nueva.java
 *
 * Created on 24 de octubre de 2007, 19:40
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cube.common.security;

import cube.common.postman.Postman;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.SecureRandom;

import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.CipherInputStream;
import javax.crypto.CipherOutputStream;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;

/**
 *
 * @author victor
 */
public class Cryptography {
    //Almacena la instancia de la clase.
    private static Cryptography myinstance;
    
    //Fichero en el que se almacenan los certificados disponibles para el nodo:
    static public final String CERT_FILE_NAME = "certKeyStore.sec";
    
    /*
     * ALGORITMOS EMPLEADOS EN LAS DISTINTAS TAREAS:
     */
    
    /**
     * Algoritmo empleado para generar numeros aleatorios.
     */
    static public final String RANDOM_ALGORITHM_NAME = "SHA1PRNG";
    /**
     * Algoritmo empleado para resumen de datos.
     */
    static public final String DIGEST_ALGORITHM_NAME = "SHA1";
    /**
     * Algoritmo empleado por la infraestructura de clave publica.
     */
    static public final String PKI_ALGORITHM_NAME = "RSA";
    /**
     * Algoritmo simetrico para cifrar bloques.
     */
    static public final String BLOCK_CIPHER_ALGORITHM_NAME = "AES";
    /**
     * Algoritmo empleado por el keyStore para almacenar las claves.
     */
    static private final String KEYSTORE_ALGORITHM_NAME = "JCEKS";
    
    
    /**
     * Tamaño del buffer empleado para la lectura/escritura de datos
     * en el cifrado por bloques
     */
    static private int LONG_BUFFER_DATA = 8;
    
    /*
     *  ALMACEN DE CERTIFICADOS EMPLADOS EN LA PLATAFORMA:
     *  <cube_path>/security
     */
    
    /**
     * Keystore que almacena los certificados empleados por Cube.
     */
    private static KeyStore certStore = null;
    /**
     * File que representa el archivo del almacen de certificados.
     */
    private static File rootCertStore = null;
    
    /**
     * Password empleado para proteger el almacen de certificados.
     */
    static private final char [] KEY_STORE_PASS_DATABASE = {'a','8','4','j','d','7','1','l','d','1'};
    
    
    //Partes estaticas:
    static 
    {
        myinstance = new Cryptography();    
    }
    
    private Cryptography() {}
    
    //Obtiene la instancia de la clase mediante el patron singleton.
    static public Cryptography getInstance(){ return myinstance;}
    
    /**
     * Abre el fichero de almacen de certificados empleado por Cube.
     * @param rootCertStore ruta del directorio en el que se almacenara el almacen de certificados.
     * @throws KeyStoreException Error general al crear el keystore.
     * @throws CertificateException Alguno de los certificados almacenados no se ha podido cargar.
     * @throws NoSuchAlgorithmException No se encuentra disponible el algoritmo empleado para comprobar
     *         la integridad del archivo de certificados.
     * @throws IOException Error debido a que el archivo puede estar corrupto.
     */
    public void openCertStore(File rootCertStore) throws KeyStoreException, CertificateException, NoSuchAlgorithmException, IOException
    {           
                certStore = KeyStore.getInstance(KEYSTORE_ALGORITHM_NAME);

                FileInputStream iStreamFileCert = null;
                this.rootCertStore = new File(rootCertStore,File.separator+CERT_FILE_NAME);
                
                if(this.rootCertStore.exists())
                {
                    try { iStreamFileCert = new FileInputStream(this.rootCertStore); } 
                    catch(Exception ex){}
                }

                certStore.load(iStreamFileCert, KEY_STORE_PASS_DATABASE);
                       
                try { if(iStreamFileCert != null) iStreamFileCert.close(); }catch(IOException ex){} 
    }
    /**
     * Vuelca el contenido del keystore al fichero de certificados.
     */
    public void saveCertStore() throws KeyStoreException, NoSuchAlgorithmException, CertificateException
    {
        try 
        {
            rootCertStore.createNewFile();
            FileOutputStream oStreamFileCert = new FileOutputStream(rootCertStore);
            
            certStore.store(oStreamFileCert, KEY_STORE_PASS_DATABASE);
            oStreamFileCert.flush();
            oStreamFileCert.close();
        } 
        
        catch (IOException ex) {
            System.err.println("Error I/O al salvar el archivo de certificados.");
        }    
    }
    
    /**
     * Añade un certificado de Autoridad Certificadora (Trusted Certificate )
     * @param alias asignado al cetificado.
     * @param Certificado. Formato empleado: <b>X509</b>.
     * @throws KeyStoreExceptio El alias ya existe pero no corresponde a una entrada tipo trusted certificate.
     *         Puede que el almacen de claves no este iniciado correctamente.
     */
    public void addCA(String alias, File certified) throws KeyStoreException, CertificateException, FileNotFoundException
    { 
        X509Certificate cert;  
        if(certStore==null) throw new KeyStoreException("KeyStore no iniciado.");
        cert = (X509Certificate) readCertificateFromFile(certified);
        certStore.setCertificateEntry(alias, cert); 
    }
    
    /**
     * Elimina una registro del almacen de certificados.
     * @throws KeyStoreException No se puede eliminar el registro o no se inicio el keystore.
     */
    public void subEntry(String alias) throws KeyStoreException
    { if(certStore!=null) certStore.deleteEntry(alias);}
    
    /**
     * Devuvelve el numero de certificados almacenados.
     * @return -1 si hay ocurrio algun problema.
     */
    public int longKeyStore() { if(certStore!=null) try { return certStore.size(); } catch (KeyStoreException ex) {} return -1;}
    
    /**
     * Devuelve los alias de los certificados almacenados.
     * @return <b>null</b> Ocurrio algun problema.
     */
    public Enumeration<String> getAliases()
    {
        if(certStore!=null)
        {
            try { return certStore.aliases();} catch (KeyStoreException ex) {}
        }
        
        return null;
    }
    
    /**
     * Determina si el alias existe en el almacen de certificados.
     */
    public boolean containAlias(String alias)
    {
        if(certStore!=null)
            try { return certStore.containsAlias(alias); } catch (KeyStoreException ex) {}
        
        return false;
    }
    
    /**
     * Obtiene el certificado referenciado por el alias.
     * @return null si el no hay almacenado un certificado con dicho alias.
     */
    public X509Certificate getCertificate(String alias)
    {
        Certificate certificado = null;
        if(certStore!=null)   
            try {
                certificado  = certStore.getCertificate(alias);
                if(certificado!=null) return (X509Certificate) certificado;
            } catch (KeyStoreException ex) {}
        
        return null;
        
    }
    
    /**
     * Indica si el certificado es tipo CA.
     * @param alias Alias del certificado.
     */
    public boolean isCAcert(String alias)
    {
        if(certStore!=null)            
            try {
                return certStore.isCertificateEntry(alias);
            } catch (KeyStoreException ex) {}
        return false;
    }
    /**
     * Obtiene un certificado tipo X.509 de un archivo empleando el estandar PKCS#7.
     * @param filename archivo del que extraer el certificado.
     * @return Certificado con formato X.509.
     * @throws CertificateException El formato del certificado no es el adecuado o esta corrupto.
     * @throws FileNotFoundException El archivo pasado como argumento no existe.
     */
    public X509Certificate readCertificateFromFile( File filename ) throws CertificateException, FileNotFoundException
    {
        FileInputStream fis = new FileInputStream(filename);
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        return (X509Certificate) cf.generateCertificate(fis);
    }
    
    /**
     * Almacena una clave privada junto con su certificado y el camino de certificacion.<br>
     * FORMATO: Certificado tipo X.509 en formato PEM. Clave privada: PKCS8.
     * @param alias Alias para la entrada.
     * @param privateFileKey Archivo con la clave privada.
     * @param certFile Archivo con el certificado asociado a la clave privada.
     * @param chain Cadena de certificacion. Se añadira automaticamente, en primer lugar, el certificado pasado como
     *        argumento.
     */
    public void addUserCertificate(String alias, File privateFileKey, File certFile,  Certificate[] chain) throws IOException, KeyStoreException, InvalidKeySpecException
    {
        if(certStore==null) throw new KeyStoreException("KeyStore no iniciado.");
        PrivateKey privKey = null;
        X509Certificate cert = null;
                
        try
        {
            privKey = readPrivateKeyFromFile(privateFileKey);
            cert = (X509Certificate) readCertificateFromFile(certFile);
        }
        catch (InvalidKeySpecException ex){throw new InvalidKeySpecException();}
        catch(Exception ex){ throw new IOException(); }     
        
        int sizeArray = 1;
        if(chain!=null) sizeArray+=chain.length;
        
        Certificate[] final_chain = new Certificate[sizeArray] ;
        final_chain[0] = cert;
        int contador = 1;
        
        if(chain!=null)
            for(Certificate certificado : chain)
            {
                final_chain[contador] = certificado;
                contador++;
            }
        
        certStore.setKeyEntry(alias, privKey, KEY_STORE_PASS_DATABASE, final_chain);
    }
    
    /**
     * Lee una clave privada de un fichero empleando el estandar PKCS#8.
     * @param privateFileKey Archivo con la clave privada.
     * @throws FileNotFoundException Archivo no encontrado.
     * @throws IOException Error I/O al leer el fichero.
     * @throws NoSuchAlgorithmException Los proveedores criptograficos instalados no soportan el algoritmo usado
     *         en la clave privada.
     * @throws InvalidKeySpecException Error al leer la clave privada. El formato no es correcto.
     */
    public PrivateKey readPrivateKeyFromFile(File privateFileKey) throws FileNotFoundException, IOException, NoSuchAlgorithmException, InvalidKeySpecException
    {
        FileInputStream fis = null;
        byte[] encodedKey = new byte[(int) privateFileKey.length()];
        
        fis = new FileInputStream(privateFileKey);
        fis.read(encodedKey);
        try { fis.close();} catch (IOException ex) {}
        
        //Crear el objeto con la clave privada:
        PKCS8EncodedKeySpec kspec = new PKCS8EncodedKeySpec(encodedKey);
        KeyFactory kfac = null;
      
        kfac = KeyFactory.getInstance(PKI_ALGORITHM_NAME);
        return kfac.generatePrivate(kspec);
    }
    
    /**
     * Genera una cadena de bytes aleatorios de tamaño numberBytes.
     * @param numberBytes numero de bytes de la cadena a generar.
     * @return null si ocurrio algun problema. De lo contrario devuelve el numero aleatorio.
     */
    static public byte [] getRandom(int numberBytes)
    {
        SecureRandom random = null;
        try {
            random = SecureRandom.getInstance(RANDOM_ALGORITHM_NAME);
        } catch (NoSuchAlgorithmException ex) {
            return null;
        }
        
         byte [] numero = new byte [numberBytes];
         random.nextBytes(numero);
         return numero;
    }
    
    /**
     * Obtiene el resumen de los datos pasados como argumento aplicando el algoritmo SHA-1
     * @param data Datos de los que obtener el resumen.
     * @throws NoSuchAlgorithmException El proveeror criptografico de Java no soporta el algoritmo empleado.
     */
    static public  byte [] digestSHA1( byte [] data ) throws NoSuchAlgorithmException
    {
        MessageDigest digest =  MessageDigest.getInstance(DIGEST_ALGORITHM_NAME);
        return digest.digest(data);
    }
    
    
    
    /**
     * Cifra un bloque de datos almacenado en una archivo. Es usado para encriptar los 
     * fragmentos de archivo.
     * @throws InvalidKeyException La clave pasada como argumento <b>no es valida</b>.
     * @throws java.io.IOException Error de lectura/escritura en los archivos pasados como argumento.
     * @throws java.security.NoSuchProviderException El proveedor de datos referenciado no se encuentra disponible.
     * @throws java.security.NoSuchAlgorithmException. El proveedor de servicios criptograficos no proporciona el algoritmo 
     *         de cifrado empleado. El nombre del algoritmo lo determina la constante <b>BLOCK_CIPHER_ALGORITHM_NAME</b>.
     * @throws NoSuchPaddingException Esquema de padding invalido o no disponible en el proveedor de servicios criptograficos.
     * @param bloqueArchivo Representa el bloque de datos a cifrar.
     * @param bloqueArchivo_cifrado Representa el bloque de datos cifrado.
     */
    static public void CipherBlock(FileInputStream bloqueArchivo, FileOutputStream bloqueArchivo_cifrado, SecretKey clave) throws NoSuchAlgorithmException, IOException, NoSuchPaddingException, InvalidKeyException
    { 
        Cipher codificador;      
        CipherInputStream cis;
        byte [] buffer_datos = new byte [LONG_BUFFER_DATA];
        int resultadoLectura = 0;
        
       
        codificador = Cipher.getInstance(BLOCK_CIPHER_ALGORITHM_NAME);      
        codificador.init(Cipher.ENCRYPT_MODE, clave);

        cis = new CipherInputStream(bloqueArchivo, codificador);

        resultadoLectura = cis.read(buffer_datos);

       while (resultadoLectura != -1) 
       {
            bloqueArchivo_cifrado.write(buffer_datos, 0, resultadoLectura);
            resultadoLectura = cis.read(buffer_datos);
       }

       bloqueArchivo_cifrado.close();
       cis.close();         
    }
    
    /**
     * Descifra un bloque de datos almacenado en una archivo. Es usado para desencriptar los 
     * fragmentos de archivo.
     * @throws InvalidKeyException La clave pasada como argumento <b>no es valida</b>.
     * @throws java.io.IOException Error de lectura/escritura en los archivos pasados como argumento.
     * @throws java.security.NoSuchProviderException El proveedor de datos referenciado no se encuentra disponible. Por defecto se emplea SunJCE.
     * @throws java.security.NoSuchAlgorithmException El proveedor de datos no soporta el algoritmo de cifrado.
     * @throws NoSuchPaddingException Esquema de padding invalido o no disponible en el proveedor de servicios criptograficos.
     * @param bloqueArchivo Representa el bloque de datos descifrado.
     * @param bloqueArchivo_cifrado Representa el bloque de datos cifrado.
     */
    static public void UnCipherBlock(FileInputStream bloqueArchivo_cifrado, FileOutputStream bloqueArchivo, SecretKey clave) throws NoSuchAlgorithmException, IOException, InvalidKeyException, NoSuchPaddingException
    {      
        Cipher codificador;      
        CipherOutputStream  cos;
        byte [] buffer_datos = new byte [LONG_BUFFER_DATA];
        int resultadoLectura = 0;
       
        codificador=Cipher.getInstance(BLOCK_CIPHER_ALGORITHM_NAME);
        codificador.init(Cipher.DECRYPT_MODE, clave);

        cos = new CipherOutputStream (bloqueArchivo, codificador);

        resultadoLectura = bloqueArchivo_cifrado.read(buffer_datos);
        while(resultadoLectura != -1)
        {
            cos.write(buffer_datos, 0, resultadoLectura);
            resultadoLectura = bloqueArchivo_cifrado.read(buffer_datos);

        }

        cos.flush();
        bloqueArchivo_cifrado.close();
        cos.close();
    }
    
    
        /**
     * Cifra un conjunto de bytes con algoritmo simetrico.
     * @return <B>true</B> datos cifrados. <B>false</B> Ocurrio un problema. Consultar el log.
     * @throws java.security.InvalidKeyException La clave proporcioada no es correcta.
     * @throws java.security.NoSuchAlgorithmException Algoritmo de cifrado No proporcionado por el conjunto de proveedores instalados en
     * Java.
     * @param clave Clave de cifrado.
     * @param data Datos origen.
     * @param destiny Resultado de la operacion.
     */
    static public boolean CipherBytes(SecretKey clave, byte data[], byte [] destiny) throws InvalidKeyException, NoSuchAlgorithmException
    {
       
        
        Cipher codificador = null;      
            
        try {
            codificador=Cipher.getInstance(BLOCK_CIPHER_ALGORITHM_NAME);
            codificador.init(Cipher.ENCRYPT_MODE, clave);
        } catch (InvalidKeyException ex) {
           
            throw new InvalidKeyException();
        } catch (NoSuchAlgorithmException ex) {
          
            throw new NoSuchAlgorithmException();
        } catch (NoSuchPaddingException ex) {
            
        }
        try {
            
            destiny = codificador.doFinal(data);
            return true;
        } catch (BadPaddingException ex) {
        } catch (IllegalBlockSizeException ex) {
           
        }
        return false;
        
    }
    
    /**
     * descifra un conjunto de bytes con algoritmo simetrico.
     * @return <B>true</B> datos cifrados. <B>false</B> Ocurrio un problema. Consultar el log.
     * @throws java.security.InvalidKeyException La clave proporcioada no es correcta.
     * @throws java.security.NoSuchAlgorithmException Algoritmo de cifrado No proporcionado por el conjunto de proveedores instalados en
     * Java.
     * @param clave Clave de cifrado.
     * @param data Datos origen.
     * @param destiny Resultado de la operacion.
     */
   static public boolean UnCipherBytes(SecretKey clave, byte data[], byte [] destiny) throws InvalidKeyException, NoSuchAlgorithmException
    {
        
        
        Cipher codificador = null;      
            
        try {
            codificador=Cipher.getInstance(BLOCK_CIPHER_ALGORITHM_NAME);
            codificador.init(Cipher.DECRYPT_MODE, clave);
        } catch (InvalidKeyException ex) {
            
            throw new InvalidKeyException();
        } catch (NoSuchAlgorithmException ex) {
            
            throw new NoSuchAlgorithmException();
        } catch (NoSuchPaddingException ex) {
            
        }
        try {
            
            destiny = codificador.doFinal(data);
            return true;
        } catch (BadPaddingException ex) {
            
        } catch (IllegalBlockSizeException ex) {}
        return false;
        
    }
}
